#include <stdlib.h>
#include <string.h>
#include <gst/gst.h>

GstElement *create_theora_decode ()
{
  static gboolean have_video = FALSE;
  GstBin *bin;
  GstElement *queue, *decoder, *cs, *sink;
  GstPad *pad;

  if (have_video) {
    g_print ("Already playing a video stream. Ignoring this one\n");
    return NULL;
  }
  
  /* create a new bin to hold the elements */
  bin = (GstBin*) gst_bin_new (NULL);
  g_assert (bin);

  /* Queue to ensure all streams can push data */
  queue = gst_element_factory_make ("queue", "q");
  g_assert (queue != NULL); /* Queue should always be available */   
  
  /* Theora decoder */
  decoder = gst_element_factory_make ("theoradec", "dec");
  if (!decoder)
    goto no_decoder;

  cs = gst_element_factory_make ("ffmpegcolorspace", "cs");
  if (!cs)
    goto no_output;
    
  /* and a video sink */
  sink = gst_element_factory_make ("autovideosink", "sink");
  if (!sink)
    goto no_output;

  /* add objects to the bin */
  gst_bin_add_many (GST_BIN (bin), queue, decoder, cs, sink, NULL);

  /* Link the elements */
  gst_element_link_many (queue, decoder, cs, sink, NULL);

  /* Retrieve the sinkpad from the queue and 'ghost' it onto the bin
   * so that the caller can find it generically */
  pad = gst_element_get_pad (queue, "sink");
  gst_element_add_pad (GST_ELEMENT (bin), gst_ghost_pad_new ("sink", pad));
  gst_object_unref (pad);

  have_video = TRUE;
  return GST_ELEMENT (bin);
no_decoder:
  g_print ("The theoradec element is not available");
  return NULL;
no_output:
  g_print ("Could not create either ffmpegcolorspace or autovideosink for output");
  return NULL;
}

GstElement *create_vorbis_decode ()
{
  static gboolean have_audio = FALSE;
  GstBin *bin;
  GstElement *queue, *decoder, *audioconvert, *sink;
  GstPad *pad;

  if (have_audio) {
    g_print ("Already playing an audio stream. Ignoring this one\n");
    return NULL;
  }

  /* create a new bin to hold the elements */
  bin = (GstBin*) gst_bin_new (NULL);
  g_assert (bin);
  
  /* Queue to ensure all streams can push data */
  queue = gst_element_factory_make ("queue", "q");
  g_assert (queue != NULL); /* Queue should always be available */
   
  /* Vorbis decoder */
  decoder = gst_element_factory_make ("vorbisdec", "dec");
  if (!decoder)
    goto no_decoder;
    
  /* an audio converter to convert vorbisdec's floating-point audio sample
   * output to integer format */
  audioconvert = gst_element_factory_make ("audioconvert", "ac");
  if (!audioconvert)
    goto no_output;
  
  /* and an audio sink */
  sink = gst_element_factory_make ("autoaudiosink", "sink");
  if (!sink)
    goto no_output;

  /* add objects to the bin */
  gst_bin_add_many (GST_BIN (bin), queue, decoder, audioconvert, sink, NULL);

  /* link the elements */
  gst_element_link_many (queue, decoder, audioconvert, sink, NULL);

  /* Retrieve the sinkpad from the queue and 'ghost' it onto the bin
   * so that the caller can find it generically */
  pad = gst_element_get_pad (queue, "sink");
  gst_element_add_pad (GST_ELEMENT (bin), gst_ghost_pad_new ("sink", pad));
  gst_object_unref (pad);

  have_audio = TRUE;
  return GST_ELEMENT (bin);
no_decoder:
  g_print ("The vorbisdec element is not available");
  return NULL;
no_output:
  g_print ("Could not create either audioconvert or autoaudiosink for output");
  return NULL;
}

void new_pad (GstElement * element, GstPad * pad, GstBin *top_pipeline)
{
  GstPad *out_pad;
  GstElement *decoder = NULL;
  GstCaps *caps;
  GstStructure *s;
  const gchar *stream_type;

  /* Decide which output we are creating based on the stream contents */
  caps = gst_pad_get_caps (pad);
  if (caps == NULL) {
    g_print ("The demuxer produced an unknown stream - ignoring\n");
    return;
  }
  
  s = gst_caps_get_structure (caps, 0);
  g_assert (s != NULL); /* Caps on a pad should always have exactly one entry */
  
  stream_type = gst_structure_get_name (s);
  
  if (strcmp (stream_type, "video/x-theora") == 0) {
    /* Is a Theora video stream */
    g_print ("Encountered a new theora video stream\n");
    decoder = create_theora_decode ();
  }
  else if (strcmp (stream_type, "audio/x-vorbis") == 0) {
    g_print ("Encountered a new vorbis audio stream\n");
    decoder = create_vorbis_decode ();
  }
  else {
    g_print ("Found unknown stream of type %s - ignoring\n", stream_type);
  }
  
  /* If no decoder was created, ignore this stream */
  if (decoder == NULL)
    return;

  /* Add the decoder into our pipeline */
  gst_bin_add (top_pipeline, decoder);
  
  /* If we created a decoder pipeline, retrieve the sink pad from it */
  out_pad = gst_element_get_pad (decoder, "sink");
  g_return_if_fail (out_pad != NULL);

  /* Attempt to link the new pad to the decoder */  
  if (gst_pad_link (pad, out_pad) != GST_PAD_LINK_OK) {
    g_print ("Failed to add the decoder for this new data stream\n");
    gst_bin_remove (top_pipeline, decoder);
    gst_object_unref (out_pad);
    return;
  }
  gst_object_unref (out_pad);
  
  /* New decoder is successfully linked in the pipeline, change its state to
   * playing so it is ready to receive data */
  gst_element_set_state (decoder, GST_STATE_PLAYING);
}

static void
event_loop (GstElement * pipe)
{
  GstBus *bus;
  GstMessage *message = NULL;

  bus = gst_element_get_bus (GST_ELEMENT (pipe));

  while (TRUE) {
    message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);

    g_assert (message != NULL);

    switch (message->type) {
      case GST_MESSAGE_EOS:
        gst_message_unref (message);
        return;
      case GST_MESSAGE_WARNING:
      case GST_MESSAGE_ERROR:{
        GError *gerror;
        gchar *debug;

        gst_message_parse_error (message, &gerror, &debug);
        gst_object_default_error (GST_MESSAGE_SRC (message), gerror, debug);
        gst_message_unref (message);
        g_error_free (gerror);
        g_free (debug);
        return;
      }
      default:
        gst_message_unref (message);
        break;
    }
  }
}

int
main (int argc, char *argv[])
{
  GstElement *bin, *filesrc, *demuxer;

  gst_init (&argc, &argv);

  if (argc != 2) {
    g_print ("usage: %s <ogg file>\n", argv[0]);
    exit (-1);
  }

  g_print ("Constructing pipeline\n");
  
  /* create a new bin to hold the elements */
  bin = gst_pipeline_new ("pipeline");
  g_assert (bin);

  /* create a disk reader */
  filesrc = gst_element_factory_make ("filesrc", "disk_source");
  g_assert (filesrc);
  g_object_set (G_OBJECT (filesrc), "location", argv[1], NULL);

  /* Create the demuxer */
  demuxer = gst_element_factory_make ("oggdemux", "demux");
  if (!demuxer) {
    g_print ("could not find element \"oggdemux\"\n");
    return -1;
  }

  /* add objects to the main pipeline */
  gst_bin_add_many (GST_BIN (bin), filesrc, demuxer, NULL);

  /* link the elements. */
  gst_element_link_many (filesrc, demuxer, NULL);

  /* Connect to the demuxer 'pad-added' signal to detect when it produces
   * a new stream */
  g_signal_connect (G_OBJECT (demuxer), "pad-added", 
      G_CALLBACK (new_pad), bin);

  g_print ("Starting playback\n");   
  /* start playing */
  gst_element_set_state (bin, GST_STATE_PLAYING);

  /* Run event loop listening for bus messages until EOS or ERROR */
  event_loop (bin);

  g_print ("Finished playback - stopping pipeline\n");
  
  /* stop the bin */
  gst_element_set_state (bin, GST_STATE_NULL);
  
  /* Unreffing the bin will clean up all its children too */
  gst_object_unref (bin);
  
  return 0;
}
