#include <stdlib.h>
#include <string.h>
#include <gst/gst.h>

GstElement *create_video_output ()
{
  static gboolean have_video = FALSE;
  GstBin *bin;
  GstElement *queue, *cs, *sink;
  GstPad *pad;

  if (have_video) {
    g_print ("Already playing a video stream. Ignoring this one\n");
    return NULL;
  }
  
  /* create a new bin to hold the elements */
  bin = (GstBin*) gst_bin_new (NULL);
  g_assert (bin);

  /* Queue to ensure all streams can push data */
  queue = gst_element_factory_make ("queue", "q");
  g_assert (queue != NULL); /* Queue should always be available */
  /* Set the queue to buffer 1/10 of a second of raw video */
	
      "max-size-bytes", 0, "max-size-buffers", 0, NULL);

  cs = gst_element_factory_make ("ffmpegcolorspace", "cs");
  if (!cs)
    goto no_output;
    
  /* and a video sink */
  sink = gst_element_factory_make ("autovideosink", "sink");
  if (!sink)
    goto no_output;

  /* add objects to the main pipeline */
  gst_bin_add_many (GST_BIN (bin), queue, cs, sink, NULL);

  /* Link the elements */
  gst_element_link_many (queue, cs, sink, NULL);

  /* Retrieve the sinkpad from the queue and 'ghost' it onto
   * the bin so that the caller can find it generically */
  pad = gst_element_get_pad (queue, "sink");
  gst_element_add_pad (GST_ELEMENT (bin), gst_ghost_pad_new ("sink", pad));
  gst_object_unref (pad);

  have_video = TRUE;
  return GST_ELEMENT (bin);
no_output:
  g_print ("Could not create either ffmpegcolorspace or autovideosink for output");
  return NULL;
}

GstElement *create_audio_output ()
{
  static gboolean have_audio = FALSE;
  GstBin *bin;
  GstElement *queue, *audioconvert, *sink;
  GstPad *pad;

  if (have_audio) {
    g_print ("Already playing an audio stream. Ignoring this one\n");
    return NULL;
  }

  /* create a new bin to hold the elements */
  bin = (GstBin*) gst_bin_new (NULL);
  g_assert (bin);
  
  /* Queue to ensure all streams can push data */
  queue = gst_element_factory_make ("queue", "q");
  g_assert (queue != NULL); /* Queue should always be available */
  /* Set the queue to buffer 1/10 of a second of raw audio */
  g_object_set (queue, "max-size-time", (GstClockTime) GST_SECOND / 10,
      "max-size-bytes", 0, "max-size-buffers", 0, NULL);

  /* an audio converter to convert floating-point audio samples to int format */
  audioconvert = gst_element_factory_make ("audioconvert", "ac");
  if (!audioconvert)
    goto no_output;
  
  /* and an audio sink */
  sink = gst_element_factory_make ("autoaudiosink", "sink");
  if (!sink)
    goto no_output;

  /* add objects to the bin */
  gst_bin_add_many (GST_BIN (bin), queue, audioconvert, sink, NULL);

  /* link the elements */
  gst_element_link_many (queue, audioconvert, sink, NULL);

  /* Retrieve the sinkpad from the queue element and 'ghost' it onto
   * the bin so that the caller can find it generically */
  pad = gst_element_get_pad (queue, "sink");
  gst_element_add_pad (GST_ELEMENT (bin), gst_ghost_pad_new ("sink", pad));
  gst_object_unref (pad);

  have_audio = TRUE;
  return GST_ELEMENT (bin);
no_output:
  g_print ("Could not create either ffmpegcolorspace or autovideosink for output");
  return NULL;
}

void new_decoded_pad (GstElement * element, GstPad * pad, gboolean last,
      GstBin *top_pipeline)
{
  GstPad *out_pad;
  GstElement *output = NULL;
  GstCaps *caps;
  GstStructure *s;
  const gchar *stream_type;

  /* Decide which output we are creating based on the stream contents */
  caps = gst_pad_get_caps (pad);
  if (caps == NULL) {
    g_print ("Decodebin produced an unknown stream - ignoring\n");
    return;
  }
  
  s = gst_caps_get_structure (caps, 0);
  g_assert (s != NULL); /* Caps on a pad should always have exactly one entry */
  
  stream_type = gst_structure_get_name (s);
  
  if (g_str_has_prefix (stream_type, "video/x-raw-")) {
    /* Is a new video stream */
    g_print ("Encountered a new video stream\n");
    output = create_video_output ();
  }
  else if (g_str_has_prefix (stream_type, "audio/x-raw-")) {
    g_print ("Encountered a new audio stream\n");
    output = create_audio_output ();
  }
  else {
    g_print ("Found unknown stream of type %s - ignoring\n", stream_type);
  }
  
  /* If no renderer was created, ignore this stream */
  if (output == NULL)
    return;

  /* Add the output into our pipeline */
  gst_bin_add (top_pipeline, output);
  
  /* If we created a output pipeline, retrieve the sink pad from it */
  out_pad = gst_element_get_pad (output, "sink");
  g_return_if_fail (out_pad != NULL);

  /* Attempt to link the new pad to the output */  
  if (gst_pad_link (pad, out_pad) != GST_PAD_LINK_OK) {
    g_print ("Failed to add the rendering pipeline for this new data stream\n");
    gst_bin_remove (top_pipeline, output);
    gst_object_unref (out_pad);
    return;
  }
  gst_object_unref (out_pad);
  
  /* New output renderer is successfully linked in the pipeline.
   * Change its state to playing so it is ready to receive data */
  gst_element_set_state (output, GST_STATE_PLAYING);
}

static void
event_loop (GstElement * pipe)
{
  GstBus *bus;
  GstMessage *message = NULL;

  bus = gst_element_get_bus (GST_ELEMENT (pipe));

  while (TRUE) {
    message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);

    g_assert (message != NULL);

    switch (message->type) {
      case GST_MESSAGE_EOS:
        gst_message_unref (message);
        return;
      case GST_MESSAGE_WARNING:
      case GST_MESSAGE_ERROR:{
        GError *gerror;
        gchar *debug;

        gst_message_parse_error (message, &gerror, &debug);
        gst_object_default_error (GST_MESSAGE_SRC (message), gerror, debug);
        gst_message_unref (message);
        g_error_free (gerror);
        g_free (debug);
        return;
      }
      default:
        gst_message_unref (message);
        break;
    }
  }
}

int
main (int argc, char *argv[])
{
  GstElement *bin, *filesrc, *decodebin;

  gst_init (&argc, &argv);

  if (argc != 2) {
    g_print ("usage: %s <file to play>\n", argv[0]);
    exit (-1);
  }

  g_print ("Constructing pipeline\n");
  
  /* create a new bin to hold the elements */
  bin = gst_pipeline_new ("pipeline");
  g_assert (bin);

  /* create a disk reader */
  filesrc = gst_element_factory_make ("filesrc", "disk_source");
  g_assert (filesrc);
  g_object_set (G_OBJECT (filesrc), "location", argv[1], NULL);

  /* Create the decodebin */
  decodebin = gst_element_factory_make ("decodebin", NULL);
  if (!decodebin) {
    g_print ("could not find the \"decodebin\" element\n");
    return -1;
  }

  /* add objects to the main pipeline */
  gst_bin_add_many (GST_BIN (bin), filesrc, decodebin, NULL);

  /* link the elements. */
  gst_element_link_many (filesrc, decodebin, NULL);

  /* Connect to decodebin's 'new-decoded-pad' signal to detect when it produces
   * a new stream */
  g_signal_connect (G_OBJECT (decodebin), "new-decoded-pad", 
      G_CALLBACK (new_decoded_pad), bin);

  g_print ("Starting playback\n");   
  /* start playing */
  gst_element_set_state (bin, GST_STATE_PLAYING);

  /* Run event loop listening for bus messages until EOS or ERROR */
  event_loop (bin);

  g_print ("Finished playback - stopping pipeline\n");
  
  /* stop the bin */
  gst_element_set_state (bin, GST_STATE_NULL);
  
  /* Unreffing the bin will clean up all its children too */
  gst_object_unref (bin);
  
  return 0;
}
